
#include "haarDecomp.h"
#include "glut.h"
#include "stdlib.h"
#include <math.h>
#include <fstream.h>

/**************************************************************************************************/
/**************************************************************************************************/

haarDecomp::haarDecomp() { 

  loaded  = false;
  image   = NULL;
  wavelet = NULL;
  currentWidth = 0;  
  currentHeight = 0;
  resX = 0;
  resY = 0;
  nbDecomposition = 0;

}

/**************************************************************************************************/
/**************************************************************************************************/

haarDecomp::~haarDecomp() {

  if (image != NULL)
    delete image;

  if (wavelet != NULL)
    delete wavelet;
	
}

/**************************************************************************************************/
/**************************************************************************************************/

void haarDecomp::setSmooth(bool s) {

  smooth = s;

}

/**************************************************************************************************/
/**************************************************************************************************/

void  haarDecomp::compress() {

  int n1,n2;
  double threshold;

  if ((pow(2,nbDecomposition)==resX)) {

    cout << "Entrer le seuil de compression : " ;
    cin >> threshold;
    cout << endl;
    
    n1=0;
    for(int i=0;i<resX*resY;i++) {
      if (wavelet[i].r==0) n1++;
      if (wavelet[i].g==0) n1++;
      if (wavelet[i].b==0) n1++;
    }
    
    cout << "Nombres de coefficients a 0 avant compression (sur " << resX*resY*3 << ") : " << n1 << endl;
    for(int i=1;i<resX*resY;i++) {
     //  if (fabs(wavelet[i].r)<threshold)
// 	wavelet[i].r=0;
//       if (fabs(wavelet[i].g)<threshold) 
// 	wavelet[i].g=0;
//       if (fabs(wavelet[i].b)<threshold) 
// 	wavelet[i].b=0;   

      if (sqrt(fabs(wavelet[i].r)+fabs(wavelet[i].g)+fabs(wavelet[i].b))<threshold) {
 	wavelet[i].r=0;  
 	wavelet[i].g=0;  
 	wavelet[i].b=0;  
      }
    }
    
    n2=0;
    for(int i=0;i<resX*resY;i++) {
      if (wavelet[i].r==0) n2++;
      if (wavelet[i].g==0) n2++;
      if (wavelet[i].b==0) n2++;
    }
    
    cout << "Nombres de coefficients a 0 apres compression (sur " << resX*resY*3 << ") : " << n2 << endl;
    cout << "Nombres de coefficients mis a 0 pendant la compression : " << n2-n1 << endl << endl;
  }

}

/**************************************************************************************************/
/**************************************************************************************************/

bool haarDecomp::reconstructionStep(struct RGB * buf,int * mark,int width,int offset) {

  //struct RGB * tmp = new (struct RGB)[width];
  struct RGB * tmp = (struct RGB*)malloc(sizeof(struct RGB)*width);

  for(int i=0;i<width/2;i++) {

    tmp[2*i].r   = (buf[i*offset].r+buf[(width/2+i)*offset].r)/M_SQRT2;
    tmp[2*i+1].r = (buf[i*offset].r-buf[(width/2+i)*offset].r)/M_SQRT2;
   
    tmp[2*i].g   = (buf[i*offset].g+buf[(width/2+i)*offset].g)/M_SQRT2;
    tmp[2*i+1].g = (buf[i*offset].g-buf[(width/2+i)*offset].g)/M_SQRT2;
    
    tmp[2*i].b   = (buf[i*offset].b+buf[(width/2+i)*offset].b)/M_SQRT2;
    tmp[2*i+1].b = (buf[i*offset].b-buf[(width/2+i)*offset].b)/M_SQRT2;

  }

  for(int i=0;i<width;i++) {
    buf[i*offset].r = tmp[i].r;
    buf[i*offset].g = tmp[i].g;
    buf[i*offset].b = tmp[i].b;
    mark[i*offset]--;
  }
   
  delete tmp;
 
 return true;

}

/**************************************************************************************************/
/**************************************************************************************************/

bool haarDecomp::reconstruct() {

  if (nbDecomposition>=1) {

    currentWidth *= 2;
    currentHeight *= 2;

    for(int col=0;col<currentWidth;col++)
      reconstructionStep(wavelet+col,mark+col,currentHeight,resX);

    for(int row=0;row<currentHeight;row++)
      reconstructionStep(wavelet+row*resX,mark+row*resX,currentWidth,1);

    nbDecomposition--;

    return true;
  }
  else 
    return false;

}

/**************************************************************************************************/
/**************************************************************************************************/

void haarDecomp:: waveletReconstruction() {
  
  while (reconstruct());

}

/**************************************************************************************************/
/**************************************************************************************************/

bool haarDecomp::decompositionStep(struct RGB * buf,int * mark,int width,int offset) {

    //struct RGB * tmp = new (struct RGB)[width];
	struct RGB * tmp = (struct RGB*)malloc(sizeof(struct RGB)*width);

  for(int i=0;i<width/2;i++) {

    tmp[i].r         = (buf[2*i*offset].r+buf[2*i*offset+offset].r)/M_SQRT2;
    tmp[width/2+i].r = (buf[2*i*offset].r-buf[2*i*offset+offset].r)/M_SQRT2;
   
    tmp[i].g         = (buf[2*i*offset].g+buf[2*i*offset+offset].g)/M_SQRT2;
    tmp[width/2+i].g = (buf[2*i*offset].g-buf[2*i*offset+offset].g)/M_SQRT2;
    
    tmp[i].b         = (buf[2*i*offset].b+buf[2*i*offset+offset].b)/M_SQRT2;
    tmp[width/2+i].b = (buf[2*i*offset].b-buf[2*i*offset+offset].b)/M_SQRT2;

  }

  for(int i=0;i<width;i++) {
    buf[i*offset].r = tmp[i].r;
    buf[i*offset].g = tmp[i].g;
    buf[i*offset].b = tmp[i].b;
    mark[i*offset]++;
  }
   
  delete tmp;
 
 return true;

}

/**************************************************************************************************/
/**************************************************************************************************/

bool haarDecomp::decompose() {
  
  if ((pow(2,nbDecomposition)<resX)) {

    for(int row=0;row<currentHeight;row++) 
      decompositionStep(wavelet+row*resX,mark+row*resX,currentWidth,1);

    for(int col=0;col<currentWidth;col++)
      decompositionStep(wavelet+col,mark+col,currentHeight,resX);

    nbDecomposition++;
    currentWidth /= 2;
    currentHeight /= 2;

    return true;
  }
  else {
    cout << "Decomposition terminee" << endl;
    return false;
  }

}

/**************************************************************************************************/
/**************************************************************************************************/

void haarDecomp::waveletDecomposition() {

  while (decompose());

}

/**************************************************************************************************/
/**************************************************************************************************/

void  haarDecomp::reset() {

  currentWidth = resX;  
  currentHeight = resY;
  nbDecomposition = 0;
 
  for(int j=0; j < resY*resX ; j++)  
    mark[j] = 0;
  
  memcpy(wavelet,image,resX * resY * sizeof(struct RGB));

  for(int j=0; j < currentWidth*currentHeight; j++) {
    wavelet[j].r /= currentWidth;
    wavelet[j].g /= currentWidth;
    wavelet[j].b /= currentWidth;
  }

}

/**************************************************************************************************/
/**************************************************************************************************/

void haarDecomp::drawImage() {

  double coef;
  int maxOffset,offset;
  double r,g,b;

  glClear( GL_COLOR_BUFFER_BIT );

  if (loaded) {
    maxOffset = resX*resY;
    glBegin(GL_POINTS);
    for(int j=0; j < resY; j++) 
      for(int i=0; i < resX; i++) {
	offset = maxOffset-resX*j-i;
	coef = resX/pow(M_SQRT2,mark[offset]);
	r = wavelet[offset].r;
	g = wavelet[offset].g;
	b = wavelet[offset].b;
	if (r<0) r = -r;
	if (g<0) g = -g;
	if (b<0) b = -b;
	glColor3f(r*coef,g*coef,b*coef);
	glVertex2f((float)i*2.0/(float)resX-1,(float)j*2.0/(float)resY-1);
      }
    glEnd();
 }
 
}

/**************************************************************************************************/
/**************************************************************************************************/

bool haarDecomp::loadImage(char name[50],int * width,int * height) {

 /* FILE* ifp;
  int ich1, ich2, maxval ;
  unsigned char idem;

  if ((ifp = fopen(name,"r"))==NULL)
    return false;

  ich1 = getc( ifp );
  if ( ich1 == EOF )
    return false;
    
  ich2 = getc( ifp );

  if ( ich2 == EOF )
    return false;

  resX = pm_getint( ifp );
  resY = pm_getint( ifp );
  maxval = pm_getint( ifp );

  // image   = new (struct RGB)[resX * resY];
  // wavelet = new (struct RGB)[resX * resY];
  image   = (struct RGB*)malloc(resX * resY*sizeof(struct RGB*));
  wavelet = (struct RGB*)malloc(resX * resY*sizeof(struct RGB*));
  mark    = new int[resX * resY];

  for(int j=0; j < resY ; j++)  
    for(int i=0; i < resX; i++) {

	 image[j * resX + (resX-i)].r = (double)pm_getrawbyte(ifp)/255.0;
	 image[j * resX + (resX-i)].g = (double)pm_getrawbyte(ifp)/255.0;
	 image[j * resX + (resX-i)].b = (double)pm_getrawbyte(ifp)/255.0;
	 mark[j*resX+i] = 0;

    }	*/ 

	ifstream ifp(name,ios::binary);
	int maxval ;
	unsigned char red,blue,green;
	char ich1, ich2;

	ifp.get(ich1);
	ifp.get(ich2);
	ifp >> resX;
	ifp >> resY;
	ifp >> maxval;

	// image   = new (struct RGB)[resX * resY];
	// wavelet = new (struct RGB)[resX * resY];
	image   = (struct RGB*)malloc(resX * resY*sizeof(struct RGB));
	wavelet = (struct RGB*)malloc(resX * resY*sizeof(struct RGB));
	mark    = new int[resX * resY];
	for(int j=0; j < resY ; j++)  
		for(int i=0; i < resX; i++) {			
			ifp.get(blue);					
			ifp.get(red);	
			ifp.get(green);						
			image[j * resX + (resX-i)].r  = (double)red/255.0;
			image[j * resX + (resX-i)].g  = (double)green/255.0;
			image[j * resX + (resX-i)].b  = (double)blue/255.0;
			mark[j*resX+i] = 0;
	}
  
	memcpy(wavelet,image,resX * resY * sizeof(struct RGB));

	*width = resX;
	*height = resY;

	currentHeight = resY;
	currentWidth = resX;

	for(int j=0; j < currentWidth*currentHeight; j++) {
		wavelet[j].r /= currentWidth;
		wavelet[j].g /= currentWidth;
		wavelet[j].b /= currentWidth;
	}

	loaded = true;  

	return loaded;

}

/**************************************************************************************************/
/**************************************************************************************************/

void haarDecomp::writePPM(int resX,int resY) {

  double r,g,b;
  ofstream of("image.ppm",ios::binary);
  
  of << "P6" ;
  of << endl ;
  of << resX << " " << resY;
  of << endl;
  of << "255";
  of << endl;

  for(int j=0;j<resY;j++) {
	for(int i=0;i<resX;i++) {
		r = wavelet[j * resX + (resX-i)].r*255.0*resX;
		g = wavelet[j * resX + (resX-i)].g*255.0*resX;
		b =	wavelet[j * resX + (resX-i)].b*255.0*resX;
		if (r>255) r = 255;
		if (g>255) g = 255;
		if (b>255) b = 255;
		of << (unsigned char)r << (unsigned char)g << (unsigned char)b;
	}
  }

  cout << "PPM writed" << endl;

}

/**************************************************************************************************/
/**************************************************************************************************/
